@@ -1,7 +1,7 @@
 /*
  * vi: formatoptions+=tc textwidth=80 tabstop=8 shiftwidth=8 noexpandtab:
  *
- * $Id: usblan.h,v 1.5 2008/04/11 05:05:26 jaymzh Exp $
+ * $Id: usblan.h,v 1.6 2013/04/10 04:17:06 jaymzh Exp $
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -23,10 +23,13 @@
 #ifndef USBLAN_H
 #define USBLAN_H
 
+#include "lc_internal.h"
+
 int InitializeUsbLan(void);
 int ShutdownUsbLan(void);
 int FindUsbLanRemote(void);
 int UsbLan_Write(unsigned int len, uint8_t *data);
 int UsbLan_Read(unsigned int &len, uint8_t *data);
+int GetXMLUserRFSetting(char **data);
 
 #endif
